@@ -6,7 +6,6 @@ from configparser import ConfigParser
 import os
 import sys
 import glob
-# from decouple import config
 import logging
 
 log = logging.getLogger('byneuron')
@@ -102,8 +101,8 @@ class ByConfig:
 
 
 class Byneuron:
-    def __init__(self, **kwargs):
-        self.config = ByConfig(**kwargs)
+    def __init__(self, api=None, login=None, key=None, secret=None):
+        self.config = ByConfig(api, login, key, secret)
         self.api = f'{self.config.url_api}/api/v1'
         self._token = ''
         self._token_expire = datetime.datetime.min.replace(tzinfo=datetime.timezone.utc)
@@ -230,15 +229,20 @@ class Byneuron:
         log.debug('_graphql %s', data)
         return self.url(url=url, data=data)
 
-    def _graphmodel(self, entities=None, events=None):
+    def _graphmodel(self, data):
         """ create / edit entities and events into the model"""
-        nodes = {e.key: e.wrap() for e in entities if isinstance(e, Entity)} \
-            if isinstance(entities, list) else {}
-        events = [e.wrap() for e in events if isinstance(e, Numberevent)] \
-            if isinstance(events, list) else []
+        entities = {}
+        numberevents = []
+        if not isinstance(data, list):
+            data = [data]
+        for d in data:
+            if isinstance(d, Entity):
+                entities.update({d.key: d.wrap()})
+            elif isinstance(d, Numberevent):
+                numberevents.append(d.wrap())
         data = {
-            "entityDataModel": {"nodes": nodes},
-            "numberEventDataModel": {"events": events}
+            "entityDataModel": {"nodes": entities},
+            "numberEventDataModel": {"events": numberevents}
         }
         log.debug('graphmodel, data: \n%s', json.dumps(data, indent=2))
         url = f'{self.api}/backend/tasks/entities/graphmodel'
@@ -257,21 +261,37 @@ class Byneuron:
             if r:
                 return r.get('publicId')
 
+    def _list_indexsetkeys(self, indexsets):
+        """
+        used for event and entity queries
+        :param indexsets:
+        :return:
+        """
+        indexsetlist = indexsets if isinstance(indexsets, list) else [indexsets]
+        indexsets = set()
+        for i in indexsetlist:
+            if isinstance(i, Entity):
+                indexsets.update({i.indexset})
+            elif isinstance(i, str):
+                indexsets.update({i})
+            elif i is None:
+                indexsets.update({self.indexset_active.key})
+        return list(indexsets)
+
     def _numberevents(self, size=100, order="DESC", deleted=False, indexsets=None, filters=None):
         """
         :param size:
         :param order:
         :param deleted:
-        :param indexsets: list of indexSet or entity keys, else active indexSet
+        :param indexsets: list of indexSetKeys, Entities, else active indexSet is set
         :param filters:
         :return:
         """
-        indexsets = [i.key if isinstance(i, Entity) else i for i in indexsets] if isinstance(indexsets, list) \
-            else [self.indexset_active.key]
         filters = [] if not isinstance(filters, list) else filters
         data = {
-            "indexSets": indexsets,  # list of indexSet keys
-            "order": order,  # "ASC","DESC"
+            "indexSets": self._list_indexsetkeys(indexsets),  # list of indexSet keys
+            "order": order,  # "ASC","DESC"  # order is respected after 'size' number of most recent events
+            # have been collected; so "DESC" gives the most consistent result
             "size": int(size),
             "deleted": bool(deleted),
             "filters": filters,
@@ -296,7 +316,6 @@ class Byneuron:
         """
 
         items = items if isinstance(items, list) else [items]
-        indexsets = list({i.indexset if isinstance(i, Entity) else self.indexset_active.key for i in items})
         timeseries = [i.key if isinstance(i, Entity) else i for i in items]
         agg_metrics = ['avg', 'min', 'max', 'count', 'diff', 'sum', 'close', 'cardinality', 'open']
         aggrs = [{
@@ -306,24 +325,24 @@ class Byneuron:
             "field": "numberValue"
         } for m in aggrs if m in agg_metrics]
         filters = filters if isinstance(filters, list) else []
-        filters.append([
-            {
-                "type": "eventitemid",
-                "values": timeseries,
-                "invertFilter": False,
-                "operator": "or"
-            },
-            {
-                "type": "range",
-                "from": int(1000 * from_date.timestamp()),
-                "to": int(1000 * to_date.timestamp()),
-                "minInclusive": True,
-                "maxInclusive": False,
-                "invertFilter": False
-            }
-        ])
+        filters.append({
+            "type": "eventitemid",
+            "values": timeseries,
+            "invertFilter": False,
+            "operator": "or"
+        })
+        filters.append({
+            "type": "range",  # this is sended as be UTC
+            "from": int(1000 * from_date.timestamp()),
+            "to": int(1000 * to_date.timestamp()),
+            "minInclusive": True,
+            "maxInclusive": False,
+            "invertFilter": False
+        })
         aggregation = {
             "type": "date",
+            "field": "timestamp",  # todo alternatives? byneuronUpdated, byneuronCreated
+            "time_zone": "UTC",  # "Africa/Nairobi" >> important on how we wrap results
             "interval": interval,
             "name": "datehistogram",
             "from": int(1000 * from_date.timestamp()), 
@@ -333,20 +352,20 @@ class Byneuron:
         }
         data = {
             "query": {
-                "indexSets": indexsets,
+                "indexSets": self._list_indexsetkeys(items),
                 "sortField": "timestamp",
-                "order": "DESC" if order is None else order, # todo order is not respected
+                "order": "ASC" if order is None else order,
                 "deleted": False if deleted is None else bool(False),
                 "filters": filters,
                 "esDocumentType": "NumberEvent"
             },
-            "gapStrategy": {} if gap is None else gap,
+            "gapStrategy": gap,
             "aggregation": aggregation
         }
-        if filters:
-            data['query']['filters'].append(filters)
+        if not isinstance(gap, dict):
+            del data['gapStrategy']
         url = f'{self.api}/backend/rest/numberevents/_datehistogramtimeline'
-        log.info('data for _datehistogramtimeline %s', data)
+        log.debug('data for _datehistogramtimeline %s', data)
         return self.url(url=url, data=data)
 
     ### endpoint handlers ###
@@ -366,12 +385,130 @@ class Byneuron:
         data = self._numberevents(indexsets=indexsets, filters=[f], size=size)
         if data:
             for i in data:
-                yield Numberevent(node=i)
+                yield Numberevent(datamodel=i)
 
     def get_numberevents_dict(self, item, size):
         return {e.datetime: e.value for e in self.get_numberevents(item=item, size=size)}
 
-    def get_aggregation(self, item, aggrs, interval, size, push=None):
+    def get_numberevents_dates(self, item, from_date=None, to_date=None, extend=None):
+        """
+        retrieve events for item as Numberevent() between two points in time
+        :param item: entity of type item
+        :param from_date: datetime to start events from (inclusive)
+        :param to_date: datetime to stop events till (exclusive)
+        :param extend: bool, add one more event before the from_date
+        :return:
+        """
+
+        if not isinstance(from_date, datetime.datetime):
+            from_date = self.now
+        if not isinstance(to_date, datetime.datetime):
+            to_date = self.now
+        order = 'DESC'
+        items = item if isinstance(item, list) else [item]
+        item_keys = list({i.key if isinstance(i, Entity) else i for i in items})
+        f_item = {
+            "type": "eventitemid",
+            "values": item_keys,
+            "invertFilter": False,
+            "operator": "or"
+        }
+        f_range = {
+            "type": "range",
+            "to": max(from_date, to_date).timestamp() * 1000,
+            "maxInclusive": False
+        }
+        if not extend:
+            f_range.update({
+                "from": min(from_date, to_date).timestamp() * 1000,
+                "minInclusive": True
+            })
+        while True:
+            data = self._numberevents(indexsets=items, order=order, filters=[f_item, f_range])
+            if data:
+                for i in data:
+                    event = Numberevent(datamodel=i)
+                    yield event
+                    if extend and event.datetime < min(from_date, to_date):
+                        return
+                f_range.update({
+                    "to": event.datetime.timestamp() * 1000,  # this is why we must have sorting to DESC
+                    "maxInclusive": False
+                })
+            else:
+                break
+
+    def get_numberevent_key(self, key, indexset=None):
+        """
+        retrieve a single event using the eventid-filter
+        :param key:
+        :param indexset:
+        :return:
+        """
+        f = {
+            "type": "eventid",
+            "values": [key],
+            "invertFilter": False,
+            "operator": "or"
+        }
+        data = self._numberevents(indexsets=indexset, filters=[f], size=1)
+        if data:
+            for i in data:
+                return Numberevent(datamodel=i)
+        data = self._numberevents(indexsets=indexset, filters=[f], size=1, deleted=True)
+        if data:
+            for i in data:
+                return Numberevent(datamodel=i)
+
+    def get_aggregation_dates(self, item, aggrs, interval, from_date, to_date, push=None, zero=None):
+        """
+        retrieve x aggregated buckets of data, including current bucket
+        :param item:
+        :param aggrs: (list of) aggregation(s) ['avg']
+        :param interval: bucket size '15m'
+        :param from_date: starting datetime
+        :param to_date: ending datetime
+        :param push: boolean to set gap strategy 'push' with extended timeline
+        :param zero: True excludes 0.0 values; False only used 0.0 values; None includes all values
+        :return:
+        """
+        aggrs = aggrs if isinstance(aggrs, list) else [aggrs]
+        gap = {"type": "push", "extendTimeline": True} if bool(push) else None
+        zerofilter = [{
+                "type": "range",
+                "name": "numberValue",
+                "from": 0.0,
+                "minInclusive": True,
+                "to": 0.0,
+                "maxInclusive": True,
+                "invertFilter": bool(zero)
+            }] if bool(zero) else None
+        data = self._datehistogramtimeline(
+            items=item,
+            aggrs=aggrs,
+            from_date=from_date,
+            to_date=to_date,
+            interval=interval,
+            gap=gap,
+            filters=zerofilter
+        )
+        if data:
+            log.debug('timings from _datehistogramtimeline: %s', {k: v for k, v in data.items() if 'Time' in k})
+            buckets = data.get('data')
+            if isinstance(buckets, list):
+                for i in buckets:
+                    ts = i.get('time')
+                    if ts:
+                        dt = datetime.datetime.fromtimestamp(ts/1000, tz=datetime.timezone.utc)
+                        # using .replace(tzinfo=utc) introduces an error due to fromtimestamp converting to local tz
+                        bucket = {
+                            'time': dt,
+                            'count': i.get('count', 0)
+                        }
+                        bucket.update(i.get('values', {}))
+                        yield bucket
+
+    def get_aggregation_size(self, item, aggrs, interval, size, push=None):
         """
         retrieve x aggregated buckets of data, including current bucket
         :param item:
@@ -381,14 +518,13 @@ class Byneuron:
         :param push: boolean to set gap strategy 'push' with extended timeline
         :return:
         """
-        aggrs = aggrs if isinstance(aggrs, list) else [aggrs]
         agg_intervals = {
             '1m': datetime.timedelta(minutes=1),
             '5m': datetime.timedelta(minutes=5),
             '15m': datetime.timedelta(minutes=15),
             '1h': datetime.timedelta(hours=1),
             '1d': datetime.timedelta(hours=24),
-            #'1M': None  # todo
+            # '1M': None  # todo
         }
 
         end = self.now
@@ -397,23 +533,15 @@ class Byneuron:
         interval_dt = agg_intervals[interval]
         end_rounded = end - datetime.timedelta(seconds=(end.timestamp() % interval_dt.total_seconds()))
         start_rounded = end_rounded - (interval_dt * (size-1))
-        gap = push if push is None else {"type": "push", "extendTimeline": True}
-        r = self._datehistogramtimeline(
-            items=item,
+        for bucket in self.get_aggregation_dates(
+            item=item,
             aggrs=aggrs,
+            interval=interval,
             from_date=start_rounded,
             to_date=end,
-            interval=interval,
-            gap=gap
-        )
-        if isinstance(r, list):
-            for i in r:
-                bucket = {
-                    'time': i.get('time', None),
-                    'count': i.get('count', 0)
-                }
-                bucket.update(i.get('values', {}))
-                yield bucket
+            push=push
+        ):
+            yield bucket
 
     def query(self, query, keys):
         """
@@ -439,7 +567,7 @@ class Byneuron:
                 result.update({k: {}})
                 for entityKey in variables.get(f"?{k}", []):
                     e = nodes.get(entityKey, {})
-                    result[k].update({entityKey: Entity(node=e)})
+                    result[k].update({entityKey: Entity(datamodel=e)})
         return result if result_with_keys else result.get(keys)
 
     def get_entities(self, entitytype, indexset=None, **kwargs):
@@ -505,32 +633,69 @@ class Byneuron:
         r = self.query(q, 'indexSet')
         return [e for e in r.values() if isinstance(e, Entity)]
 
+    def datamodel(self, data):
+        data = self._graphmodel(data)
+        if data:
+            log.debug('graphmodel returned %s', data)
+            for k, v in data.get('result', {}).items():
+                events = v.get('events', [])
+                nodes = v.get('nodes', {})
+                if k == 'entityDataModel':
+                    for i in nodes.values():
+                        yield Entity(datamodel=i)
+                elif k == 'numberEventDataModel':
+                    for i in events:
+                        yield Numberevent(datamodel=i)
+
     def create_entity(self, entity):
         log.info('create_entity %s', entity)
         entity.set_context('create')
         log.debug('create_entity %s %s %s', entity.entity_type, entity.role, entity.type)
         if not entity.entity_type == 'Gateway':
-
             # indexSetCode only for entity that are defined in indexSet
             if entity.role:
                 for role in entity.role:
                     self.set_isc(f'{entity.entity_type.lower()}Role', role)
             if entity.type:
                 self.set_isc(f'{entity.entity_type.lower()}Type', entity.type)
-
-        data = self._graphmodel(entities=[entity])
-        if data:
-            log.debug('created entity data %s', data)
-            result = data.get('result', {}).get('entityDataModel', {}).get('nodes', {}).get(entity.key, {})
-            log.info('created entity %s', result)
-            return Entity(node=result)
+        for i in self.datamodel(entity):
+            if i.key == entity.key:
+                log.info('created entity %s', i)
+                return i
 
     def update_entity(self, entity):
-        entity.set_context('update')
-        data = self._graphmodel(entities=[entity])
-        if data:
-            result = data.get('result', {}).get('entityDataModel', {}).get('nodes', {}).get(entity.key, {})
-            return Entity(node=result)
+        if isinstance(entity, Entity):
+            for i in self.update(entity):
+                return i
+
+    def delete_entity(self, entity, undo=False):
+        if isinstance(entity, Entity):
+            for i in self.update(entity, delete=not undo):
+                return i
+
+    def update_event(self, event):
+        if isinstance(event, Numberevent):
+            for i in self.update(event):
+                return i
+
+    def delete_event(self, event, undo=False):
+        if isinstance(event, Numberevent):
+            for i in self.update(event, delete=not undo):
+                return i
+
+    def update(self, datamodel, delete=None):
+        datamodels = datamodel if isinstance(datamodel, list) else [datamodel]
+        for d in datamodels:
+            if isinstance(d, (Numberevent, Entity)):
+                if delete is True:
+                    d.set_context('delete')  # <- method to delete;
+                    # d._datamodel.update({'deleted': True})  # <- does not work!
+                elif delete is False:
+                    d.set_context('undelete')
+                    #d._datamodel.update({'deleted': False})  # method to undo deletion
+                else:
+                    d.set_context('update')
+        return [i for i in self.datamodel(datamodels)]
 
     def create_item(self, **kwargs):
 
@@ -660,8 +825,8 @@ class Byneuron:
 
 
 class Entity:
-    def __init__(self, node=None):
-        self._datamodel = {} if not isinstance(node, dict) else node
+    def __init__(self, datamodel=None):
+        self._datamodel = {} if not isinstance(datamodel, dict) else datamodel
 
     def __repr__(self):
         return f'{self.entity_type} {self.name}'
@@ -721,6 +886,10 @@ class Entity:
     def description(self):
         return self._prop('description')
 
+    @property
+    def deleted(self):
+        return self._prop('deleted')
+
     @property
     def name(self):
         return self._attr('name')
@@ -808,6 +977,8 @@ class Entity:
 
     @property
     def indexset(self):
+        if self.entity_type == 'IndexSet':
+            return self.key
         if self.entity_type == 'Gateway':
             return self._link('isAssignedTo')
         return self._link('isDefinedIn')
@@ -860,10 +1031,9 @@ class Entity:
         return self
 
     def set_context(self, action):
-        if 'entityContext' not in self._datamodel:
-            self._datamodel.update({'entityContext': {}})
-        if 'action' not in self._datamodel['entityContext']:
-            self._datamodel['entityContext'].update({"action": action})
+        context = self._datamodel.get('entityContext', {})
+        context.update({"action": action})
+        self._datamodel.update({"entityContext": context})
 
     def set_link(self, linktype, entity):
         def _index():
@@ -975,16 +1145,14 @@ class Entity:
 
 class Numberevent:
 
-    def __init__(self, node=None):
-        if isinstance(node, dict):
-            for k, v in node.items():
-                setattr(self, k, v)
+    def __init__(self, datamodel=None):
+        self._datamodel = datamodel if isinstance(datamodel, dict) else {}
 
     def __repr__(self):
-        return f'Numberevent {self.datetime} {self.value}'
+        return f'Numberevent {self.datetime} {self.value}{", deleted" if self.deleted else ""}'
 
     def wrap(self):
-        return vars(self)
+        return self._datamodel
 
     def create(self, value, dt, item):
         """
@@ -1001,7 +1169,7 @@ class Numberevent:
                 dt.replace(tzinfo=datetime.timezone.utc)
             return dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
 
-        node = {
+        datamodel = {
             "timestamp": timestamp(),
             "id": str(uuid.uuid4()),  # todo is this needed
             "itemId": item.key,
@@ -1012,24 +1180,69 @@ class Numberevent:
             "indexSetKey": item.indexset,
             "numberValue": float(value)
         }
-        for key, value in node.items():
-            setattr(self, key, value)
+        self._datamodel = datamodel
+
+    def set_context(self, action):
+        context = self._datamodel.get('eventContext', {})
+        context.update({"action": action})
+        self._datamodel.update({"eventContext":context})
+    def set_tag(self, tag):
+        if 'tags' not in self._datamodel:
+            self._datamodel.update({'tags': [tag]})
+        elif tag not in self.tags:
+            self._datamodel['tags'].append(tag)
+
+    @staticmethod
+    def to_datetime(timestamp):
+        if timestamp is not None:
+            return datetime.datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S.%f%z')
 
     @property
     def datetime(self):
-        if hasattr(self, 'timestamp'):
-            return datetime.datetime.strptime(self.timestamp, '%Y-%m-%dT%H:%M:%S.%f%z')
+        return self.to_datetime(self._datamodel.get('timestamp'))
 
     @property
     def value(self):
-        if hasattr(self, 'numberValue'):
-            return float(self.numberValue)
+        value = self._datamodel.get('numberValue')
+        if value is not None:
+            return value  # float(value)
 
     @property
     def now(self):
         """timezone aware utc now"""
         return datetime.datetime.now(datetime.timezone.utc)
 
+    @property
+    def deleted(self):
+        return self._datamodel.get('deleted')
+
+    @property
+    def id(self):  # this is the id of the event
+        return self._datamodel.get('id')
+
+    @property
+    def item_id(self):  # this is the id of the item linking to this event
+        return self._datamodel.get('itemId')
+
+    @property
+    def tags(self):
+        return self._datamodel.get('tags', [])
+
+    @property
+    def dates(self):
+        dates = self._datamodel.get('dates')
+        if dates is not None:
+            return {i.get('type'): self.to_datetime(i.get('timestamp')) for i in dates}
+        return {}
+
+    @property
+    def datetime_created(self):
+        return self.dates.get('byneuronCreated')
+
+    @property
+    def datetime_updated(self):
+        return self.dates.get('byneuronUpdated')
+
 
 def example01():
     logging.basicConfig(level=logging.INFO)
@@ -1047,5 +1260,23 @@ def example02():
         r.create_item(name='test')
 
 
+def example03():
+    logging.basicConfig(level=logging.INFO)
+    log.info('get event')
+    # crete item 'test' in indexSet 0
+    with Byneuron() as r:
+        e = r.get_numberevent_key(
+            key='b3875301-23ae-46ce-bf11-53f53a9bceb8',
+            indexset='7710db50-7088-11ed-93aa-db27429c6866')
+        log.info('found event %s: %s', e, e._datamodel)
+
+        if e.deleted:
+            e = r.delete_event(e, undo=True)
+            log.info('undeleted example event %s', e)
+        else:
+            e = r.delete_event(e)
+            log.info('deleted example event %s', e)
+
+
 if __name__ == '__main__':
     example01()
